from __future__ import print_function
from builtins import str
from builtins import range
from builtins import object
import logging
import os
from collections import OrderedDict

from raven import Client
from raven.conf import setup_logging
from raven.handlers.logging import SentryHandler
from raven.transport.http import HTTPTransport

from sentrylogger import credentials as creds
from sentrylogger import sentry_constants as sntry_cnsts


'''
def trace_logger(frame, event, arg):
    """
    Adds support for logging function calls to Sentry.

    Logs every function call present in the program structure
    to Sentry with arguments.

    Parameters
    ----------
    frame: frame
        The stack frame from the code being run. Consists all the data
        associated to the function such as arguments, name, etc.
    event: String
        A string naming the type of notification. Example: call, return, etc.
    arg: Object
        An event specific value such as the value being returned, exception or
        None.
    """
    global log_funcs
    log_msg = func_validity(frame, event, arg, log_funcs)
    if log_msg:
        sentryClient().queue_debug_log(log_msg[0], data=log_msg[1])
'''


def ecs_error_logger(s_client, err_msg, failures=None):
    """
    Logs ECS Errors to Sentry.

    If a failure has occured, logs an error to Sentry and logs the DEBUG
    queue to Sentry.

    Parameters
    ----------
    failures: String
        List of failures.
    s_client: sentryClient object
        Used to log.
    err_msg: String
        The error message to log.
    """
    if failures:
        s_client.log_error(err_msg)
        s_client.queue_debug_log(err_msg,
                                 data=OrderedDict(enumerate(failures)))
        s_client.logall_debug_queue()


class sentryClient(object):
    """
    Wrapper class to make sentryClient singleton.
    """
    class __sentryClient(object):
        """
        Represents a client for logging through sentry.
        Uses SENTRY_DSN for setting up the client from environment with
        stack trace enabled and sets log level to DEBUG.
        Provides methods for logging at various levels to sentry:
            1) log_info: To log at INFO level with a message. This log contains
                only the message.
            2) queue_debug_log: Adds a DEBUG log to the debuglist (size 80).
                Old logs are deleted and not sent to Sentry when debuglist gets
                full.
                DEBUG log contains the message and optional data.
            3) log_error: To log at ERROR level with a message. This log
                contains the message and program stack.
            4) logall_debug_queue: Sends all debuglist logs to Sentry.
        """

        def __init__(self):
            """
            Sets up Sentry logger and client.

            Initializes Sentry client, handler and logger using
            SENTRY_DSN variable from Environment.
            """
            # Setup System Log objects
            self.logger = logging.getLogger(__name__)
            self.logger.setLevel(logging.DEBUG)
            
            # Setup Handlers and client params
            if os.getenv('LAMBDA_TASK_ROOT'):
                print("Initializing Lambda SentryClient")
                self.client = Client(creds.SENTRY_DSN, extra=None,
                                     enable_breadcrumbs=False,
                                     transport=HTTPTransport)
            else:
                print("Initializing Async SentryClient")
                self.client = Client(creds.SENTRY_DSN, extra=None,
                                     enable_breadcrumbs=False)
            self.handler = SentryHandler(self.client)
            self.handler.setLevel(logging.DEBUG)
            setup_logging(self.handler)
            self.debuglist = []

    sentry_client = None

    def __init__(self):
        if not sentryClient.sentry_client:
            sentryClient.sentry_client = sentryClient.__sentryClient()

    def __getattr__(self):
        return getattr(self.sentry_client)

    def log_info(self, message):
        """
        Logs at Info Level with a message.

        Parameters
        ----------
        message: String
            An info message to log.
        """
        self.sentry_client.logger.info(message)

    def log_error(self, message):
        """
        Logs at Error Level with a message.

        Parameters
        ----------
        message: String
            An info message to log.
        """
        self.sentry_client.logger.error(message, exc_info=True)

    def queue_debug_log(self, message, data=None):
        """
        Queues the debug log to debuglist, according to size of extra data.

        Queues the message for sending to Sentry with a limit of 80 messages.
        Older messages are popped out when queue gets full and only 80 are
        maintained in the queue. These 80 messages are logged to Sentry only
        when an error occurs or the queue is flushed using the
        logall_debug_queue method. Messages with extra data size <= 500B are
        added to the debug_queue. Messages with extra data size > 500B are
        split into separate log messages and added to the queue while appending
        information such as PART x/y for every message.

        Parameters
        ----------
        message: String
            A debug message to log.
        data: Dictionary/String (Optional)
            A data dictionary/string with extra data to log.
        """
        message = str(message)
        if data is None:
            if sntry_cnsts.FUNCTION_CALL not in message:
                message = sntry_cnsts.USER_DEBUG + message
            if len(self.sentry_client.debuglist) >= 80:
                    self.sentry_client.debuglist.pop(0)
            if message in self.sentry_client.debuglist:
                    message = message + sntry_cnsts.PLUS_ONE
            self.sentry_client.debuglist.append((message, data))
        elif len(data) <= sntry_cnsts.MAX_DATA_SIZE:
            if sntry_cnsts.FUNCTION_CALL not in message:
                message = sntry_cnsts.USER_DEBUG + message
            if len(self.sentry_client.debuglist) >= 80:
                    self.sentry_client.debuglist.pop(0)
            if message in self.sentry_client.debuglist:
                    message = message + sntry_cnsts.PLUS_ONE
            self.sentry_client.debuglist.append((message, data))
        else:
            data = str(data)
            split_data = {message + str(sntry_cnsts.PART +
                          str(i)): data[i: i+sntry_cnsts.MAX_DATA_SIZE]
                          for i in
                          range(0, len(data), sntry_cnsts.MAX_DATA_SIZE)}
            for k, v in split_data.items():
                self.queue_debug_log(k, v)

    def logall_debug_queue(self):
        """
        Sends the logs in debuglist queue to Sentry in one message.
        """
        data = OrderedDict(self.sentry_client.debuglist)
        self.sentry_client.logger.debug(sntry_cnsts.ALL_DEBUG_LOGS, extra={
            'data': data,
        })
        del self.sentry_client.debuglist[:]
